@@ -24,9 +24,8 @@ import pandas as pd
 import numpy as np
 from statsmodels.graphics.gofplots import qqplot
 from scipy.optimize import curve_fit
-
-# from itertools import chain
-# import csv
+import scipy.odr as sco
+import FlowCal
 
 standardColorsHex = ['#5B84B1FF', '#FC766AFF', '#5F4B8BFF', '#E69A8DFF',
                      '#42EADDFF', '#CDB599FF', '#00A4CCFF', '#F95700FF',
@@ -262,8 +261,7 @@ def plot_data(p, xs, ys, ttl=None, dlab=None, xlab=None, ylab=None, ms=None, lw=
 
 def string_to_float(potential_float):
     """
-    Converts string to float if possible (that is unless ValueError is
-                                          encountered).
+    Converts string to float if possible (that is unless ValueError is encountered).
 
     Parameters
     ----------
@@ -283,154 +281,153 @@ def string_to_float(potential_float):
         return potential_float
 
 
-def file_select(path, set_cols=None, cut_rows=None, separator=None, py_axlist=True,
-                as_matrix=False):
+def file_select(path, set_cols=None, cut_rows=None, **kwargs):
     """
-    This function selects and extracts data, from a file at a specified path.
-    It can be useful to index multiple data files in a way, that allows for
-    easy extration in a for-loop.
+    This function selects and extracts data, from a file at a specified path. It can be useful to index multiple data 
+    files in a way, that allows for easy extration in a for-loop.
 
     Parameters
-    ----------
-    path : string
-        Defines the file path, note that you might want to do this as an
-        r-string (and if for-loop; part as an f-string).
-    set_cols : list of ints, optional
-        List of the column indexes you want extracted
-        (note that this is not a range, but specific selection).
-        The default is [0,1].
-    cut_rows : int or list, optional
-        If integer; cut from row 0 to specified integer, if list; cut the
-        specified rows from the list. The default is 0.
-    separator : string, optional
-        Define the deliminter of the data set (if nescessary). The default is
-        if .csv; \',\', if .txt; \'\\t\'.
-    py_axlist : bool, optional
-        Constructs a regular python list, consisting of lists of all values of
-        a certian variable, instead of gaining rows of value-sets.
-        The default is False.
-    as_matrix : bool, optional
-        Allows for loading of data as a matrix via numpy.loadtxt; note that
-        this is only valid for .txt files. The default is False.
+        path : string
+            Defines the file path, note that you might want to do this as an r-string (and if for-loop; part as an 
+            f-string).
+        set_cols : list of ints, optional
+            List of the column indexes you want extracted (note that this is not a range, but specific selection). The 
+            default is [0,1].
+        cut_rows : int or list, optional
+            If integer; cut from row 0 to specified integer, if list; cut the specified rows from the list. The default 
+            is 0.
+
+    Keyword Arguments
+        separator : string, optional
+            Define the deliminter of the data set (if nescessary). The default is if .csv; \',\', if .txt; \'\\t\'.
+        py_axlist : bool, optional
+            Constructs a regular python list, consisting of lists of all values of a certian variable, instead of 
+            gaining rows of value-sets. The default is False.
+        as_matrix : bool, optional
+            Allows for loading of data as a matrix via numpy.loadtxt; note that this is only valid for .txt files. The 
+            default is False.
 
     Returns
-    -------
-    data : list
-        List (or list of lists) with the data from the selected file under the
-        specified conditions.
-    data_axlist : list
-        Instead of containing data points from the data set, contains what
-        corresponds to an x-, y-, z- etc. lists. Only relavant if
-        py_axlist = True; then the function yields both data and data_axlist.
+        data : list
+            List (or list of lists) with the data from the selected file under the specified conditions.
+        data_axlist : list
+            Instead of containing data points from the data set, contains what corresponds to an x-, y-, z- etc. lists. 
+            Only relavant if py_axlist = True; then the function yields both data and data_axlist.
 
     """
-    assert path, 'No path selected.'
-    if not set_cols:
-        set_cols_fixed = [0, 1]
-    if isinstance(set_cols, int):
-        set_cols_fixed = [set_cols]
+    assert path, 'Missing file path.'  # assert if missing path
+    if not set_cols:  # define standard column selection if no costum selection is defined
+        set_columns = [0, 1]
     else:
-        set_cols_fixed = set_cols
+        if isinstance(set_cols, int):
+            set_columns = [set_cols]
+        else:
+            set_columns = set_cols
     # if not cut_rows:
     #     try:
-    #         cut_rows = 1
+    #         cut_rows = 0
     #     except ValueError:
     #         while ValueError:
-    # cut_rows =+1
-    allowed_extensions = ['.csv', '.txt', '.excel', '.xlsx', '.dat']
-    file_extension = os.path.splitext(path)[1]
+    #               cut_rows += 1
 
+    # define list of passable extensions
+    allowed_extensions = ('.csv', '.txt', '.excel', '.xlsx', '.dat', '.fcs')
+    file_extension = os.path.splitext(path)[1]  # split input file name and save extension
+
+    # check if passed extension can be handled
     if file_extension not in allowed_extensions:
-        raise ValueError(f'Selected file type {file_extension} is invalid.')
+        raise ValueError(f'Selected file type {file_extension} is not supported.')
+
     # try to define standard delimiter, if none is defined
-    if not separator:
+    if 'separator' not in kwargs.keys():
         if file_extension == '.csv':
             separator = ','
         elif file_extension in ('.txt', '.dat'):
-            if as_matrix is True:
+            if 'as_matrix' in kwargs.keys() and kwargs.get('as_matrix'):
                 separator = None
-            elif as_matrix is False:
+            else:
                 separator = '\t'
+    else:
+        separator = kwargs.get('separator')
     if file_extension in ('.excel', '.xlsx'):
-        data = pd.read_excel(path, header=cut_rows, usecols=set_cols_fixed
-                             ).to_numpy()
+        data = pd.read_excel(path, header=cut_rows, usecols=set_columns).to_numpy()
     elif file_extension in ('.csv', '.txt', '.dat'):
-        if as_matrix is True:
+        if 'as_matrix' in kwargs.keys() and kwargs.get('as_matrix'):
             data = np.loadtxt(fname=path, delimiter=separator, skiprows=cut_rows)
-        elif as_matrix is False:
-            data = pd.read_csv(path, header=cut_rows, usecols=set_cols_fixed,
-                               sep=separator).to_numpy()
-    if py_axlist is True:
+        else:
+            data = pd.read_csv(path, header=cut_rows, usecols=set_columns, sep=separator).to_numpy()
+    elif file_extension in ('.fcs'):
+        data = FlowCal.io.FCSData(path)
+    if 'py_axlist' not in kwargs.keys() or ('py_axlist' in kwargs.keys() and kwargs.get('py_axlist')):
         data_axlist = [data[:, i].tolist() for i in range(len(data[0]))]
         data_axlist_fix = [[string_to_float(i) for i in data_axlist[j]]
                            for j in range(len(data_axlist))]
         result = data_axlist_fix
-    if py_axlist is False:
+    else:
         result = data
     return result
 
 
-def fit_data(function, x_list, y_list, g_list, rel_var=False, N=None, mxf=None,
-             extMin=None, extMax=None, **kwargs):
+def fit_data(func, x_list, y_list, g_list, method='curve_fit', **kwargs):
     """
     Fits data to the given general function, and outputs the parameters for
     the specific function.
 
     Parameters
-    ----------
-    function : lambda/py function
-        The specific function data is to be fitted to
-    x_list : list
-        x-list data.
-    y_list : list
-        y-list data.
-    g_list : list
-        Guess-list. These are initial guesses at the parameters in the
-        function to fit.
-    rel_var : bool, optional
-        Determines whether the calculated variation should be relative or
-        absolute.
-        The default is False.
-    N : int, optional
-        The number of constructed data-points. The default is the size of the
-        provided x-list.
-    mxf : int, optional
-        The maximum amount of iterations. The default is 1000.
-    extMin : float, optional
-        Extrapolate data points down to this limit.
-    extMax : float, optional
-        Extrapolate data points up to this limit.
+        func : function
+            The specific function data is to be fitted to
+        x_list : list
+            x-list data.
+        y_list : list
+            y-list data.
+        g_list : list
+            Guess-list. These are initial guesses at the parameters in the
+            function to fit.
+        method : str, optional
+            Specific method of fitting. Currently, options are curve_fit and odr. The default is curve_fit.
+
+    Keyword Arguments
+        f_num : int
+            The number of constructed data-points. The default is 300.
+        mxf : int
+            The maximum amount of iterations. The default is 1000.
+        extrp : float, int or list
+            Extrapolate fitted x and y lists. If a value is given, it is determined whether it is a minimum or maximum
+            extrapolation, if list, the first element will be minimum and the second element the maximum.
 
     Returns
-    -------
-    popt : list
-        Fitted parameters in the same order as defined in the provided
-        function.
-    pcov_fix : list
-        The covariance for the determined parameters.
-    pstd : lsit
-        The standard deviation of the determined parameters.
-    xs_fit : list
-        List of fitted x-values.
-    ys_fit : list
-        List of fitted y-values.
+        popt : list
+            Fitted parameters in the same order as defined in the provided
+            function.
+        pcov : list
+            The covariance for the determined parameters.
+        pstd : list
+            The standard deviation of the determined parameters.
+        x_fit : list
+            Fitted x-values.
+        y_fit : list
+            Fitted y-values.
+        x_err_est : list, conditional
+            Estimated input x-errors if odr is the method.
+        y_err_est : list, conditional
+            Estimated input y-errors if odr is the method.
 
     """
 
-    # define allowed kwargs
-    kwargs_list = ('rel_var', 'N', 'mxf', 'extrp')
-    assert all(kwargs.keys()) not in kwargs_list, 'Passed kwarg is undefined.'
+    # send warning if more than 15 different constants in function
+    if len(g_list) > 15:
+        warnings.warn('Fitting more than 15 constants may take a while.')
 
-    if not N:
-        N = len(x_list)
-    if not mxf:
+    # define standard params if none is set
+    if 'f_num' in kwargs.keys():
+        frame_number = kwargs.get('f_num')
+    else:
+        frame_number = 300
+    if 'mxf' in kwargs.keys():
+        mxf = kwargs.get('mxf')
+    else:
         mxf = 1000
 
-    popt, pcov = curve_fit(f=function, xdata=x_list, ydata=y_list, p0=g_list,
-                           absolute_sigma=rel_var, maxfev=mxf)
-    pcov_fix = [pcov[i][i] for i in range(len(popt))]
-    pstd = [np.sqrt(i) for i in pcov_fix]
-
     x_min_temp = min(x_list)
     x_max_temp = max(x_list)
 
@@ -453,21 +450,47 @@ def fit_data(function, x_list, y_list, g_list, rel_var=False, N=None, mxf=None,
             raise ValueError(
                 'Extrapolation must be of type int, float or list.')
     else:
-        if not extMin:
-            x_min = x_min_temp
-        else:
-            x_min = extMin
-        if not extMax:
-            x_max = x_max_temp
-        else:
-            x_max = extMax
+        x_min = x_min_temp
+        x_max = x_max_temp
 
-    xs_fit = np.linspace(x_min, x_max, N)
-    ys_fit = function(xs_fit, *[i for i in popt])
+    # define x-fit list from defined x min and max values
+    x_fit = np.linspace(x_min, x_max, frame_number)
 
-    if len(popt) > 15:
-        warnings.warn('Fitting more than 15 constants may take a while.')
-    return list(popt), pcov_fix, pstd, xs_fit, ys_fit
+    # perform fitting after given method
+    if method == 'curve_fit':
+        popt, pcov = curve_fit(f=func, xdata=x_list, ydata=y_list, p0=g_list, absolute_sigma=True, maxfev=mxf)
+        pstd = np.sqrt(np.diag(pcov))
+
+        # define output
+        y_fit = func(x_fit, *popt)  # define y-list
+        __out__ = (popt, pcov, pstd, x_fit, y_fit)
+    elif method == 'odr':
+        if 'x_err' in kwargs.keys():  # check for given x error
+            x_err = kwargs.get('x_err')
+        else:
+            x_err = None
+        if 'y_err' in kwargs.keys():
+            y_err = kwargs.get('y_err')
+        else:
+            y_err = None
+        odr_fit_function = sco.Model(func)  # define odr model
+        odr_data = sco.RealData(x_list, y_list, sx=x_err, sy=y_err)  # define odr data
+        odr_setup = sco.ODR(odr_data, odr_fit_function, beta0=g_list)  # define the ODR itself
+        odr_out = odr_setup.run()  # run the ODR
+
+        # define constants, along with covariance and deviation
+        popt, pcov, pstd = odr_out.beta, odr_out.cov_beta, odr_out.sd_beta
+
+        # define estimated x and y errors
+        x_err_est, y_err_est = odr_out.delta, odr_out.eps
+
+        # define output
+        y_fit = func(popt, x_fit)
+        __out__ = (popt, pcov, pstd, x_fit, y_fit, x_err_est, y_err_est)
+    else:
+        raise ValueError(f'Passed method, {method}, is not supported.')
+
+    return __out__
 
 
 def step_finder(x_data, y_data, delta=30, lin=0.005, err=0.005):
@@ -502,16 +525,14 @@ def step_finder(x_data, y_data, delta=30, lin=0.005, err=0.005):
     final_point = initial_point + delta
     x_test, y_test = x_data[initial_point:final_point], y_data[initial_point:
                                                                final_point]
-    popt, pcov_fix, pstd, xs_fit, ys_fit = fit_data(linear_fit, x_test, y_test,
-                                                    [0, 1])
+    popt, pcov_fix, pstd, xs_fit, ys_fit = fit_data(linear_fit, x_test, y_test, [0, 1])
     xs_point, ys_point = [], []
     while final_point < len(x_data):
         initial_point += 1
         final_point = initial_point + delta
         x_test, y_test = x_data[initial_point:final_point], y_data[
                                                             initial_point:final_point]
-        popt, pcov_fix, pstd, xs_fit, ys_fit = fit_data(linear_fit, x_test,
-                                                        y_test, [0, 1])
+        popt, pcov_fix, pstd, xs_fit, ys_fit = fit_data(linear_fit, x_test, y_test, [0, 1])
         if abs(popt[0]) < lin and pstd[0] < err:
             xs_point.append(sts.mean(xs_fit))
             ys_point.append(sts.mean(ys_fit))
